function bound = bnds(Omega,epsilon,aOmega,areps)
%
%  Compute the m-point finite frequency Kramers-Kronig
%  relation.  See Milton, Eyre and Mantese.
%
%  Usage:
%
%    bound = bnds(Omega,epsilon,aOmega,areps)
%
%  On entry:
%
%    Omega   -- frequencies of imaginary part of eps
%    epsilon -- the dielectric constant
%        The first column of eps must be the imaginary part
%        The second column is optional.  If it is included,
%        it must be the real part of the dielectric
%        constant that has been computed by KK from the
%        imaginary part.  The calculation should be
%        done with the imaginary part set to zero outside
%        the given frequency range.
%    aOmega -- frequencies of the anchors
%    areps  -- anchors, real part of the dielectric constant
%
%  On return:
%
%    bound(:,1) -- lower m-point bound at frequencies Omega
%    bound(:,2) -- upper m-point bound at frequencies Omega
%

%
%  extract the real and imaginary parts
%
esizec = size(epsilon(1,:));
ieps = epsilon(:,1);
if esizec ==2 
  kkreps = epsilon(:,2);
end

%
%  rescale the frequencies
%
num_data = length(Omega);
omega = Omega/Omega(num_data);
zed = omega.^2;
z_minus = zed(1);
z_plus = zed(num_data);

%
%  interpolate to regular spacing
%
step = (zed(num_data)-zed(1))/(2^(nextpow2(num_data))-1);
z = (zed(1):step:zed(num_data))';
iepsi = interp1(zed,ieps,z,'linear');
num_z = length(z);
%
%  Hilbert tranform in new variables to calculate real 
%  eps, and interpolate back to original spacing
%
x = [iepsi;zeros(num_z,1)];
y = -imag(hilbert(x));
kkrepsz = y(1:num_z);
if esizec ==1
  kkreps = interp1(z,kkrepsz,zed,'linear');
end

%
%  data anchors
%
num_anchors = length(aOmega);
azed = (aOmega/Omega(num_data)).^2;
aepkk = areps - interp1(zed,kkreps,azed,'linear');
azedi = length(zed(zed<azed(1)));

%
%  one point upper/lower bound
%
fb11 = aepkk(1)*ones(azedi,1);
fb12 = azed(1)*aepkk(1)./zed(1:azedi);
fb1 = min([fb11';fb12'])';
fb21 = aepkk(1)*ones(num_data-azedi,1);
fb22 = azed(1)*aepkk(1)./zed(azedi+1:num_data);
fb1(azedi+1:num_data) = max([fb21';fb22'])';
 
%
%  check anchors against the one point bounds
%
if num_anchors > 1
  bcheck = interp1(zed,fb1,azed(2),'linear');
  if azed(2) > azed(1) & aepkk(2) < bcheck
    'one point bound violated at second anchor'
    return
  elseif azed(2) < azed(1) & aepkk(2) > bcheck
    'one point bound violated at second anchor'
    return
  end

%
%  one point bound
%
elseif num_anchors == 1
  bound(1:azedi,1) = fb1(1:azedi);
  bound(azedi+1:num_data,1) = fb1(azedi+1:num_data);
end

%
%  m-point bounds  (m>=2)
%
npol = [1 3 3 1]';
z1 = [0 zed(1) zed(num_data)]';
z2 = [0 0 zed(1); zed(1) zed(num_data) zed(num_data)];

%
% loop over the number of bounds, 
%   clear the previous bounds
%
for m=2:num_anchors
  kflag = 0;
  clear f_low_m;
  clear f_up_m;
  fbmm = zeros(num_data,8); jj=0;
%
%  loop over the number prescribed poles
%
  for k=0:min([m,3])
%
%  compute the number of free zeros and poles
%
    alpha = floor((m+k-1)/2);
    beta  = floor((m-k)/2);
%
%  anchors to be used
%
    anchzed = azed(1:m);
    ancheps = aepkk(1:m);
%
%  loop over the possible prescribed poles
%
    for j=1:npol(k+1)
%
%  prescribed poles
%
      if k == 0
        ppol = [];
      elseif k == 1
        ppol = z1(j);
      elseif k == 2
        ppol = z2(:,j);
      elseif k==3
        ppol = z1;
      end
%
%  Using a power basis for the numerator and
%    denominator, interpolate the anchors.
%    CAUTION: This is a Hilbert like matrix
%
      Q = ones(m,1);
      for kpol=1:k
        Q = Q.*(anchzed-ppol(kpol));
      end
      fQ = ancheps.*Q;
      A = zeros(m);
      for n=0:alpha
        A(:,n+1) = anchzed.^(alpha-n);
      end
      for n=0:beta-1
        A(:,n+alpha+2) = -fQ.*(anchzed.^(beta-n-1));
      end
%
%  Power basis coefficients of numerator and denominator
%
      b = A\(fQ.*(anchzed.^beta));
      nump = b(1:alpha+1);
      denp = [1; b(alpha+2:m)];
%
%  compute and sort the poles and zeros
%
      pol = sort([roots(denp);ppol]);
      zer = sort(roots(nump));
%
%  interlace the poles and zeros -- pole first, then
%    zero, then pole, etc...
%
      lace(1:2:2*length(pol)) = pol;
      lace(2:2:2*length(zer)+1) = zer;
%
%  check non-negativity of poles and zeros
%
      fbm_ok = 0;
      if min(lace) < 0 
        fbm_ok = fbm_ok + 1;
      end
%
%  check for interlace of poles and zeros
%
      if max(abs(lace-sort(lace))) > 0
        fbm_ok = fbm_ok + 1;
      end
%
%  check for all poles outside of (z_minus, z_plus)
%
      px = pol(pol<z_plus);
      if length( px(px>z_minus) ) > 0
        fbm_ok = fbm_ok + 1;
      end
%
%  if analytic constraints are satisfied, evaluate candidate 
%
      if fbm_ok == 0
%
%  evaluate power basis polynomials just computed
%
        numer = polyval(nump,zed);
        denom = polyval(denp,zed);
        'possible resonances'
        pol
%
%  evaluate prescribed polynomial in denominator
%
        for kpol = 1:k
          denom = denom.*(zed-ppol(kpol));
        end
%
%  calculate the candidate
%
        denom(denom==0) = eps*ones(length(denom(denom==0)),1);
        fbm  = (numer./denom)+kkreps;
        jj = jj+1;
        fbmm(:,jj) = fbm;
%
%  update the bound
%
        if kflag == 0
          f_low_m = fbm;
          f_up_m  = fbm;
          kflag = 1;
        elseif kflag > 0
          f_low_m = min([f_low_m'; fbm'])';
          f_up_m  = max([f_up_m'; fbm'])';
          kflag = kflag+1;
        end
      end
%
%  close loop on possible prescribed poles (j)
%
    end
%
%  close loop on number of prescribed poles (k)
%
  end
%
%  evalulate the bounds if they have been computed
%
  if kflag >= 2

    bound(:,1) = f_low_m;
    bound(:,2) = f_up_m;
%
%  check for compatibitity at the next point
%
    if m < num_anchors
      anchzed = azed(alpha+beta+2);
      ancheps = aepkk(alpha+beta+2);
      lbchk = interp1(zed,f_low_m,anchzed,'linear');
      ubchk = interp1(zed,f_up_m,anchzed,'linear');
      if ancheps > lbchk & ancheps < ubchk
        m
        'm point bound violated at m+1 st anchor'
        return
      end
    end
%
%  bounds not computed, so outta here
%
  else
    m
    'Analytic contraints violated on m-point bound'
    return
  end
%
%  close loop on bound (m)
%
end
